import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;

public class ImageBuilder {
	// this is not only used for the X symbol but also the base block size. Keep quadratic
	private static final String[] X_SYMBOL = new String[]{
		"          ",
		" #      # ",
		"  #    #  ",
		"   #  #   ",
		"    ##    ",
		"    ##    ",
		"   #  #   ",
		"  #    #  ",
		" #      # ",
		"          "};
	private static final BufferedImage X_IMAGE = new BufferedImage(X_SYMBOL[0].length(), X_SYMBOL.length, BufferedImage.TYPE_INT_RGB);
	private static final BufferedImage X_IMAGE_ERROR = new BufferedImage(X_SYMBOL[0].length(), X_SYMBOL.length, BufferedImage.TYPE_INT_RGB);
	static {
		for (int x = 0; x < X_SYMBOL[0].length(); ++x) {
			for (int y = 0; y < X_SYMBOL.length; ++y) {
				Color color = X_SYMBOL[y].charAt(x) == '#' ? CellType.MARKED_AS_EMPTY.getColor() : CellType.EMPTY.getColor();
				X_IMAGE.setRGB(x, y, color.getRGB());
				color = X_SYMBOL[y].charAt(x) == '#' ? CellType.ERRORENOUS_MARKED_AS_FILLED.getColor() : CellType.ERRORENOUS_MARKED_AS_EMPTY.getColor();
				X_IMAGE_ERROR.setRGB(x, y, color.getRGB());
			}
		}
	}
	private static final int BLOCK_SIDE_LENGTH = X_SYMBOL.length;
	private static final int LINE_WIDTH = 1;
	private static final int WIDER_LINE_WIDTH = 2;
	
	public static BufferedImage createPixelImage(CellType[][] cellTypes, float zoom) {
		int width = cellTypes.length;
		int height = cellTypes[0].length;
		BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		for (int x = 0; x < width; x++) {
		    for (int y = 0; y < height; y++) {
		    	Color color = cellTypes[x][y].getColor();
		        bufferedImage.setRGB(x, y, color.getRGB());
		    }
		}
		
		return zoomImage(bufferedImage, zoom);
	}
	
	private static BufferedImage zoomImage(BufferedImage image, float zoomRatio) {
		int w = image.getWidth();
		int h = image.getHeight();
		BufferedImage after = new BufferedImage((int)(w * zoomRatio), (int)(h * zoomRatio), BufferedImage.TYPE_INT_ARGB);
		AffineTransform at = new AffineTransform();
		at.scale(zoomRatio, zoomRatio);
		AffineTransformOp scaleOp = 
		   new AffineTransformOp(at, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
		after = scaleOp.filter(image, after);
		return after;
	}
	
	public static BufferedImage createGridImage(CellType[][] cellTypes) {
		int widthPixel = cellTypes.length;
		int heightPixel = cellTypes[0].length;
		
		// transformed pos on grid of last px plus one block
		int width = calculatePositionOnGrid((widthPixel - 1), BLOCK_SIDE_LENGTH, LINE_WIDTH, WIDER_LINE_WIDTH) + BLOCK_SIDE_LENGTH;
		int height = calculatePositionOnGrid((heightPixel - 1), BLOCK_SIDE_LENGTH, LINE_WIDTH, WIDER_LINE_WIDTH) + BLOCK_SIDE_LENGTH;
		
		BufferedImage bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
		Graphics2D graphics = bufferedImage.createGraphics();
		graphics.setPaint(CellType.EMPTY.getColor());
		graphics.fillRect(0, 0, width, height);
		
		// add pixels
		for (int x = 0; x < widthPixel; x++) {
		    for (int y = 0; y < heightPixel; y++) {
		    	CellType type = cellTypes[x][y];
		    	int targetx = calculatePositionOnGrid(x, BLOCK_SIDE_LENGTH, LINE_WIDTH, WIDER_LINE_WIDTH);
		    	int targety = calculatePositionOnGrid(y, BLOCK_SIDE_LENGTH, LINE_WIDTH, WIDER_LINE_WIDTH);
		    	switch(type) {
					case ERRORENOUS_MARKED_AS_EMPTY:
						graphics.drawImage(X_IMAGE_ERROR, targetx, targety, null);
						break;
//					case MARKED:
//						break;
					case MARKED_AS_EMPTY:
						graphics.drawImage(X_IMAGE, targetx, targety, null);
						break;
					default:
						graphics.setPaint(type.getColor());
						graphics.fillRect(targetx, targety, BLOCK_SIDE_LENGTH, BLOCK_SIDE_LENGTH);
						break;
		    	}
		    }
		}
		
		// add grid lines
		graphics.setPaint(Color.GRAY);
		for (int x = 0; x < widthPixel; x++) {
			// probably a better way to do this lol
			int targetx = calculatePositionOnGrid(x, BLOCK_SIDE_LENGTH, LINE_WIDTH, WIDER_LINE_WIDTH) + BLOCK_SIDE_LENGTH;
			if ((x + 1) % 5 == 0) {
				// wide line
				graphics.fillRect(targetx, 0, WIDER_LINE_WIDTH, height);
			} else {
				graphics.fillRect(targetx, 0, LINE_WIDTH, height);
			}
		}
		for (int y = 0; y < heightPixel; y++) {
			// probably a better way to do this lol
			int targety = calculatePositionOnGrid(y, BLOCK_SIDE_LENGTH, LINE_WIDTH, WIDER_LINE_WIDTH) + BLOCK_SIDE_LENGTH;
			
			if ((y + 1) % 5 == 0) {
				// wide line
				graphics.fillRect(0, targety, width, WIDER_LINE_WIDTH);
			} else {
				graphics.fillRect(0, targety, width, LINE_WIDTH);
			}
		}
		graphics.dispose();
		
		
		return bufferedImage;
	}
	
	private static int calculatePositionOnGrid(int pixelPosition, int blockSideLength, int gridlineWidth, int wideGridlineWidth) {
		// for a start, assume there are no grid lines. As index start with 0, this is simple
		int targetPosition = pixelPosition * blockSideLength;
		
		// grid has one line after every block and a wider line after every 5. lets do the latter first
		int wideGridLinesToPosition = (int)Math.floor(pixelPosition/5);
		// add just the difference to normal lines, so we can ignore wide lines for normal lines
		targetPosition += (wideGridLinesToPosition * (wideGridlineWidth - gridlineWidth));
		
		// and one line after every block.
		targetPosition += (pixelPosition * gridlineWidth);
		
		return targetPosition;
	}
	
	public static int getGridLengthFromPixelLength(int pixelLength) {
		return calculatePositionOnGrid((pixelLength - 1), BLOCK_SIDE_LENGTH, LINE_WIDTH, WIDER_LINE_WIDTH) + BLOCK_SIDE_LENGTH;
	}
	
	
}